using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

/// <summary>
/// This plugin is about body transparency and hiding body parts, while also maintaining functionality of other plugins that manipulate skin material shaders (e.g., tesselation).
/// According to legend, there once was a plugin "Character Transparency" by ZRSX based on Doofenschmalphys edit of Seamonkey's "Hide character" plugin, which itself is a modification of Acidbubbles' "Improved POV".
/// 
/// Version history:
///   0.1: trying to figure out how this works
///   0.2: adding groups
///   0.3: adding selection of parameter to tweak
///   0.4: bugfixes around change of skin
///   0.5: adding special render queues
///   0.6: adding visibility toggles
///   1.0: producing first .var
///   1.1: fixing eye reflection bug
/// </summary>

namespace HAL9001
{
	public class GoingGoingGone : MVRScript
	{
		#region PluginInfo    
		
			public string pluginAuthor      = "HAL9001";
			public string pluginName        = "Going Going Gone";
			public string pluginVersion     = "1.1";
			public string pluginDate        = "2023-04-08";
			public string pluginDescription = "Makes body parts transparent or hides them completely";
			
		#endregion

		#region Vars  

			protected Atom                 m_person;
			protected DAZCharacter         m_character;
			protected DAZCharacterSelector m_characterSelector;
			protected DAZSkinV2            m_activeSkin;
		
			protected Dictionary<string, SkinPartControl>  m_controls;
			protected Dictionary<string, SkinGroupControl> m_groupControls;
			
			protected JSONStorableFloat         m_globalTransparency;
			protected JSONStorableFloat         m_globalRenderQueue;
			protected JSONStorableStringChooser m_parameterToEdit;
			protected JSONStorableBool          m_globalDebugOutput;

		#endregion
		
		protected const float DEFAULT_TRANSPARENCY = 0.0f;
		protected const int   DEFAULT_RENDER_QUEUE = 2450;
		
		protected const string PARAMETER_VISIBILITY   = "Visibility";
		protected const string PARAMETER_TRANSPARENCY = "Transparency";
		protected const string PARAMETER_RENDERQUEUE  = "Render Queue";

		protected static readonly string[] BodyMaterialNames = new[]
		{
			"G:Head:Back of the head, face, ears, nostrils & mouth",
				"Head|Back of Head",
				"Face",
				"Ears",
				"Eyelashes",
				"Irises",
				"Cornea",
				"Pupils",
				"Sclera",
				"EyeReflection|Eye Reflection",
				"Lacrimals",
				"Tear",
				"Nostrils",
				"Lips",
				"Gums",
				"Teeth",
				"Tongue",
				"InnerMouth|Inner Mouth",

			"G:Upper Body:Neck, shoulder, torso & arms",
				"Neck",
				"Shoulders",
				"Torso",
				"Nipples",
				"Forearms",
				"Hands",
				"Fingernails",

			"G:Lower Body:Hips, genitals & legs",
				"Hips",
				"defaultMat|Genitals Female",
				"Genitalia|Genitals Male",
				"Genitals",
				"Anus|Anus Male",
				"Legs",
				"Feet",
				"Toenails",

			// "Hidden",
		};

		protected static readonly Dictionary<string, int> SpecialPartsRenderQueueDefault = new Dictionary<string, int>
		{
			{ "Tear", 3000 }
		};

		protected static readonly Dictionary<string, Shader> TransparencyReplacementShaders = new Dictionary<string, Shader>
		{
			// Opaque materials
			{ "Custom/Subsurface/GlossCullComputeBuff",              Shader.Find("Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/GlossNMCullComputeBuff",            Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/GlossNMDetailCullComputeBuff",      Shader.Find("Custom/Subsurface/TransparentGlossNMDetailNoCullSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/CullComputeBuff",                   Shader.Find("Custom/Subsurface/TransparentSeparateAlphaComputeBuff") },
			{ "Custom/Subsurface/GlossNMTessMappedFixedComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
			// Transparent materials (don't need a replacement, but can be "driven" with "AlphaAdjust")
			{ "Custom/Subsurface/TransparentComputeBuff",                                 null },
			{ "Custom/Subsurface/TransparentSeparateAlphaComputeBuff",                    null },
			{ "Custom/Subsurface/TransparentGlossComputeBuff",                            null },
			{ "Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff",               null },
			{ "Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff",             null },
			{ "Custom/Subsurface/TransparentGlossNMDetailNoCullSeparateAlphaComputeBuff", null },
			{ "Custom/Subsurface/TransparentGlossNoCullSeparateAlphaComputeBuff",         null },
			{ "Custom/Subsurface/AlphaMaskComputeBuff",                                   null },
			{ "Marmoset/Transparent/Simple Glass/Specular IBLComputeBuff",                null },
		};


		protected class SkinPartControl
		{
			protected GoingGoingGone     parentScript;
			protected string             storableName;
			protected JSONStorableBool   visibilityToggle;
			protected JSONStorableFloat  currTransparency;
			protected JSONStorableFloat  currRenderQueue;
			
			protected float    oldTransparency;
			protected int      oldRenderQueue;

			protected Material material;
			protected Shader   originalShader, replacementShader;
			protected float    originalAlphaAdjust;
			protected Color    originalColor;
			protected Color    originalSpecular;
			protected int      originalQueue;
			protected bool     replacementShaderActive;

			public SkinPartControl(GoingGoingGone _parent, string _storableName)
			{
				parentScript = _parent;
				storableName = _storableName;

				visibilityToggle = new JSONStorableBool(storableName + " Hidden", false);
				parentScript.RegisterBool(visibilityToggle);

				currTransparency = new JSONStorableFloat(storableName + " Transparency", DEFAULT_TRANSPARENCY, 0f, 1f, true);
				parentScript.RegisterFloat(currTransparency);
				
				int defaultRQ;
				if (!SpecialPartsRenderQueueDefault.TryGetValue(storableName, out defaultRQ)) { defaultRQ = DEFAULT_RENDER_QUEUE; }
				currRenderQueue = new JSONStorableFloat(storableName + " Render Queue", defaultRQ, 0f, 5000f, true);
				parentScript.RegisterFloat(currRenderQueue);
				
				oldTransparency   = -1.0f; // force update ASAP
				material          = null;
				originalShader    = null;
				replacementShader = null;

				replacementShaderActive = false;
			}

			public void CreateUI()
			{
				switch (parentScript.m_parameterToEdit.val)
				{
					case PARAMETER_VISIBILITY :
					{
						var tglT = parentScript.CreateToggle(visibilityToggle, false);
						tglT.label = storableName;
						break;
					}

					case PARAMETER_TRANSPARENCY : 
					{
						var sldT = parentScript.CreateSlider(currTransparency, false);
						sldT.label              = storableName;
						sldT.rangeAdjustEnabled = false;
						break;
					}
					case PARAMETER_RENDERQUEUE :
					{
						var sldRQ = parentScript.CreateSlider(currRenderQueue, false);
						sldRQ.valueFormat = "F0";
						break;
					}
				}
			}

			public void DestroyUI()
			{
				parentScript.RemoveToggle(visibilityToggle);
				parentScript.RemoveSlider(currTransparency);
				parentScript.RemoveSlider(currRenderQueue);
			}

			public void Reset()
			{
				if (replacementShaderActive) RestoreShader();
				material                = null;
				originalShader          = null;
				replacementShader       = null;
				replacementShaderActive = false;
			}

			public void SetMaterial(Material _material)
			{
				if ((material == null) && (_material != null))
				{
					material                = _material;
					originalShader          = material.shader;
					replacementShaderActive = false;
					// SuperController.LogMessage(String.Format("Skin part '{0}' controlled", material.name));
				}
				else
				{
					// some materials occur several times, e.g., with hip.
					// just ignore this
					
					// SuperController.LogMessage(String.Format("Skin part '{0}' already controlled", material.name));
				}
			}

			public bool ApplyTransparency()
			{
				if (material == null) return false;

				bool  shaderReplaced = false;
				bool  shaderChanged  = replacementShaderActive && (material.shader != replacementShader); // something changed the replacement shader externally
				float transparency   = visibilityToggle.val ? 1.0f : currTransparency.val;                // "hidden" parameter overrides transparency
				int   renderQueue    = (int) currRenderQueue.val;
				if (shaderChanged || (transparency != oldTransparency) || (renderQueue != oldRenderQueue))
				{
					if (transparency <= 0.0f)
					{
						// no transparency > restore original shaders
						if (replacementShaderActive)
						{
							shaderReplaced |= RestoreShader();
						}
					}
					else
					{
						if (!replacementShaderActive || shaderChanged)
						{
							// transparency needed (or shader changed) > shader replacement necessary
							shaderReplaced |= ReplaceShader();
						}
					}

					if (replacementShaderActive)
					{
						// shader was replaced > set render queue and alpha
						material.renderQueue = renderQueue;
						material.SetFloat("_AlphaAdjust", Mathf.Min(originalAlphaAdjust, -transparency));
						Color col = originalColor;
						col.a = 1.0f - transparency;
						material.SetColor("_Color", col);
						col    = originalSpecular;
						col.a  = 1.0f - transparency;
						if (storableName == "Eye Reflection")
						{
							// treat differently to avoid reflection showing up for too long
							col.r *= col.a;
							col.g *= col.a;
							col.b *= col.a;
						}
						material.SetColor("_SpecColor", col);
					}

					oldTransparency = transparency;
					oldRenderQueue  = renderQueue;
				}
				return shaderReplaced;
			}

			public bool ReplaceShader()
			{
				bool shaderReplaced = false;
				if (material != null)
				{
					if (TransparencyReplacementShaders.TryGetValue(material.shader.name, out replacementShader))
					{
						originalShader      = material.shader;
						originalAlphaAdjust = material.GetFloat("_AlphaAdjust");
						originalColor       = material.GetColor("_Color");
						originalSpecular    = material.GetColor("_SpecColor");
						originalQueue       = material.renderQueue;
						// parentScript.log("Render Queue for " +storableName + " = " + originalQueue);

						if (replacementShader != null)
						{
							material.shader         = replacementShader;
							replacementShaderActive = true;
							shaderReplaced          = true;
						
							parentScript.log(String.Format("Replaced shader for material {0} ({1})", material.name, replacementShader.name));
						}
						else
						{
							// "null" as replacement shader means that the material already has a transparent capable shader
							// "keep" original shader
							replacementShader       = material.shader;
							replacementShaderActive = true;
						}
					}
				}
				return shaderReplaced;
			}

			public bool RestoreShader()
			{
				bool shaderRestored = false;
				if ((material != null) && replacementShaderActive)
				{
					material.renderQueue = originalQueue;
					material.shader      = originalShader;
					material.SetFloat("_AlphaAdjust", originalAlphaAdjust);
					material.SetColor("_Color",       originalColor);
					material.SetColor("_SpecColor",   originalSpecular);
					
					if (replacementShader != originalShader)
					{
						parentScript.log(String.Format("Restored shader for material {0} ({1})", material.name, originalShader.name));
					}

					originalShader          = null; // restore has happened
					replacementShader       = null;
					replacementShaderActive = false;
					shaderRestored          = true;
				}
				return shaderRestored;
			}

			public void SetVisibility(bool _hidden)
			{
				visibilityToggle.val = _hidden;
			}

			public void SetTransparency(float _transparency)
			{
				currTransparency.val = _transparency;
			}

			public void SetRenderQueue(int _renderQueue)
			{
				currRenderQueue.val = _renderQueue;
			}

			public void ForceUpdate()
			{
				oldTransparency = -1;
			}
		}


		protected class SkinGroupControl
		{
			protected GoingGoingGone     parentScript;
			protected string             groupName;

			protected JSONStorableString groupHeader;
			protected JSONStorableFloat  groupTransparency;
			protected JSONStorableBool   visibilityToggle;
			protected JSONStorableBool   groupExpanded;

			protected List<SkinPartControl> controls;

			public SkinGroupControl(GoingGoingGone _parent, string _groupName, string _description)
			{
				parentScript = _parent;
				groupName    = _groupName;

				groupHeader = new JSONStorableString(groupName, "<color=#000><size=32><b>" + groupName + "</b></size></color>\n" + _description);

				groupExpanded = new JSONStorableBool("Expand Group", false, (bool _val) => parentScript.RefreshUI());

				groupTransparency = new JSONStorableFloat(_groupName + " Group Transparency", 0f, (_val) => ApplyGroupTransparency(_val), 0f, 1f, true);
				parentScript.RegisterFloat(groupTransparency);
				groupTransparency.isStorable   = false;
				groupTransparency.isRestorable = false;

				visibilityToggle = new JSONStorableBool(_groupName + " Group Hidden", false, (bool _val) => SetVisibility(_val));
				parentScript.RegisterBool(visibilityToggle);
				visibilityToggle.isStorable   = false;
				visibilityToggle.isRestorable = false;
				
				controls = new List<SkinPartControl>();
			}

			public void AddControl(SkinPartControl _control)
			{
				controls.Add(_control);
			}

			public void CreateUI()
			{
				var txtGroupHeader = parentScript.CreateTextField(groupHeader, false);
				txtGroupHeader.height = 50.0f;

				var tglExpand = parentScript.CreateToggle(groupExpanded, false);

				var sldT   = parentScript.CreateSlider(groupTransparency, false);
				sldT.label = "Group Transparency";

				if (groupExpanded.val == true)
				{
					foreach (SkinPartControl control in controls) 
					{
						control.CreateUI();
					}
				}
			}

			public void DestroyUI()
			{
				foreach (SkinPartControl control in controls) 
				{
					control.DestroyUI();
				}
				parentScript.RemoveSlider(groupTransparency);
				parentScript.RemoveToggle(groupExpanded);
				parentScript.RemoveTextField(groupHeader);
			}

			public void SetVisibility(bool _hidden)
			{
				foreach (var control in controls)
				{
					control.SetVisibility(_hidden);
				}
			}

			public void SetTransparency(float _value)
			{
				groupTransparency.val = _value;
			}

			protected void ApplyGroupTransparency(float _value)
			{
				foreach (var control in controls)
				{
					control.SetTransparency(_value);
				}
			}
		}

		
		public override void Init()
		{
			try
			{
				if (containingAtom?.type != "Person")
				{
					SuperController.LogError($"This plugin only works on Person atoms");
					DestroyImmediate(this);
					return;
				}

				m_person            = containingAtom;
				m_characterSelector = m_person.GetComponentInChildren<DAZCharacterSelector>();
				m_character         = null;
				m_activeSkin        = null;

				// Help text
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=32><b>Help</b></size></color>\n\n" +
					"<color=#333>" +
					"This plugin changes the visibility and transparency of body parts.\n\n" +
					"Moving the sliders will change the transparency of the related body part.\n\n" +
					"For quickly applying the same transparency to all body parts, use the 'Global Transparency' slider (be aware that it will change all sliders).\n\n" +
					"</color>");
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 500.0f;

				m_parameterToEdit = new JSONStorableStringChooser("Parameter to Edit", 
					new List<string>{ PARAMETER_VISIBILITY, PARAMETER_TRANSPARENCY, PARAMETER_RENDERQUEUE },
					PARAMETER_VISIBILITY, "Parameter to Edit",
					(string _val) => RefreshUI()
				);
				m_parameterToEdit.isStorable = false;
				CreatePopup(m_parameterToEdit, true);

				m_globalTransparency = new JSONStorableFloat("Global Transparency", DEFAULT_TRANSPARENCY, (_val) => ApplyTransparencyToAll(_val), 0f, 1f, true);
				RegisterFloat(m_globalTransparency);
				m_globalTransparency.isStorable   = false;
				m_globalTransparency.isRestorable = false;
				var sliderT = CreateSlider(m_globalTransparency, true);

				m_globalRenderQueue = new JSONStorableFloat("Render Queue", DEFAULT_RENDER_QUEUE, (_val) => ApplyRenderQueueToAll(_val), 0, 5000, true);
				RegisterFloat(m_globalRenderQueue);
				m_globalRenderQueue.isStorable   = false;
				m_globalRenderQueue.isRestorable = false;
				var sliderRQ = CreateSlider(m_globalRenderQueue, true);
				sliderRQ.valueFormat = "F0";
				
				m_globalDebugOutput = new JSONStorableBool("Debug Output", false);
				RegisterBool(m_globalDebugOutput);
				var tglDebug = CreateToggle(m_globalDebugOutput, true);

				m_controls      = new Dictionary<string, SkinPartControl>();
				m_groupControls = new Dictionary<string, SkinGroupControl>();
				SkinGroupControl activeGroup = null;
				foreach (string bodyMaterialName in BodyMaterialNames)
				{
					if (bodyMaterialName.StartsWith("G:"))
					{
						// control group
						string[] parts       = bodyMaterialName.Split(':');
						string   title       = parts[1];
						string   description = parts[2];
						
						activeGroup = new SkinGroupControl(this, title, description);
						m_groupControls[title] = activeGroup;
					}
					else
					{
						// "normal" part
						string[] parts        = bodyMaterialName.Split('|');
						string   storableName = (parts.Count() > 1) ? parts[1] : parts[0];
						string   materialName = parts[0];
						
						var control = new SkinPartControl(this, storableName);
						m_controls[materialName] = control;
						activeGroup.AddControl(control);
					}					
				}

				RefreshUI();
			}
			catch (Exception _cause)
			{
				logError("Initialisation error " + _cause);
				DestroyImmediate(this);
			}
		}

		public void RefreshUI()
		{
			// brute force approach > delete all controls on the left
			foreach (var group in m_groupControls)
			{
				group.Value.DestroyUI();
			}
			// then recreate them 
			foreach (var group in m_groupControls)
			{
				group.Value.CreateUI();
			}
		}

		protected void ApplyTransparencyToAll(float _value)
		{
			foreach (var group in m_groupControls)
			{
				group.Value.SetTransparency(_value);
			}
		}

		protected void ApplyRenderQueueToAll(float _value)
		{
			foreach (var control in m_controls)
			{
				control.Value.SetRenderQueue((int) _value);
			}
		}

		public void OnEnable()
		{
			ForceUpdate();
		}
		
		public void OnDisable()
		{
			RestoreShaders();
		}
		
		public void OnDestroy()
		{
			OnDisable();
		}

		public void Update()
		{
			if (m_characterSelector.selectedCharacter != m_character)
			{
				RestoreShaders();
				m_activeSkin = null;
				m_character  = m_characterSelector.selectedCharacter;
			}

			if (m_character != null)
			{
				bool shaderChanged = false;
				if (m_activeSkin != m_character.skin)
				{
					log("Skin change detected");
					
					// skin changed > reset controls
					foreach (var control in m_controls) 
					{
						control.Value.Reset();
					}
					
					m_activeSkin = m_character.skin;
					foreach (var material in m_activeSkin.GPUmaterials)
					{
						string partName = material.name.Split('-')[0]; // remove "-1/2/3..." postfix used for multiple characters
						SkinPartControl control;
						if (m_controls.TryGetValue(partName, out control))
						{
							control.SetMaterial(material);
						}
						else
						{
							logError(String.Format("No skin part controller for material '{0}'", material.name));
						}
					}

					ForceUpdate();
					shaderChanged = true;
				}
					
				foreach (var control in m_controls)
				{
					shaderChanged |= control.Value.ApplyTransparency();
				}
				if (shaderChanged)
				{
					// Hack to force a refresh of the shaders cache
					m_activeSkin?.BroadcastMessage("OnApplicationFocus", true);
				}
			}
		}

		protected void ForceUpdate()
		{
			foreach (var control in m_controls)
			{
				control.Value.ForceUpdate();
			}
		}

		protected void RestoreShaders()
		{
			foreach (var control in m_controls)
			{
				control.Value.RestoreShader();
			}
			// Hack to force a refresh of the shaders cache
			m_activeSkin?.BroadcastMessage("OnApplicationFocus", true);
		}


		protected void log(String _msg)
		{
			if (m_globalDebugOutput.val == true)
			{
				SuperController.LogMessage("GoingGoingGone (" + m_person.name + "): " + _msg);
			}
		}

		protected void logError(String _msg)
		{
			SuperController.LogError("GoingGoingGone (" + m_person.name + "): " + _msg);
		}
	}
}